### A Pluto.jl notebook ###
# v0.19.13

using Markdown
using InteractiveUtils

# ╔═╡ d99395ed-2bb6-4998-8aed-e288f73b031b
begin
	using CSV, DataFrames, PlutoUI, Random, Statistics
	using LinearAlgebra: norm
	using StatsBase
	using Distributions
	using MLDatasets
end

# ╔═╡ 013e24a0-2787-11ec-1249-5f47f0e4e6a2
begin
	_check_complete(complete) = complete ? "✅" : "❌"

	md"""
	# Setup

	This section loads and install all the packages. You should be setup already from Assignment 1, but if not please read and follow the `instructions.md` for further details.
	"""
end

# ╔═╡ d3b72240-b983-4dfa-bbea-c5ada977a4a9
PlutoUI.TableOfContents()

# ╔═╡ 4c285c36-a7f4-4c85-8b80-8444fc7e0f30
begin
	function isclose(a, b, prec=1.e-8)
		norm(a - b) < prec
	end
end

# ╔═╡ 9c1a3c78-8094-4d34-81e4-5b1d68df0d8d
md"""
### !!!IMPORTANT!!!

Insert your details below. You should see a green checkmark.
"""

# ╔═╡ ac6c7a54-5480-458c-8eb6-3fca2c415263
student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)

# ╔═╡ eb7624a9-334b-4c57-8fac-9c000a217b99
let
	def_student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)
	if length(keys(def_student) ∩ keys(student)) != length(keys(def_student))
		md"You don't have all the right entries! Make sure you have `name`, `email`, `ccid`, `idnumber`. ❌"
	elseif any(getfield(def_student, k) == getfield(student, k) for k in keys(def_student))
		md"You haven't filled in all your details! ❌"
	elseif !all(typeof(getfield(def_student, k)) === typeof(getfield(student, k)) for k in keys(def_student))
		md"Your types seem to be off: `name::String`, `email::String`, `ccid::String`, `idnumber::Int`"
	else
		md"Welcome $(student.name)! ✅"
	end
end

# ╔═╡ 69312dd9-4cda-4983-850e-f4f9222b89d5
md"""
# Toy Data

Let's first start by importing a dataset for use throughout the assignment.
Later in the assignment, we will be using the full dataset for evaluating our models.
But for debugging purposes, we will start with a small subset of the data to test our models very quickly.
"""

# ╔═╡ 0a308ac8-18b6-44aa-a27e-86358a1a89b5
Random.seed!(367)

# ╔═╡ 86056b72-ba93-4a53-9943-bc7bc9ceeb12
function oneHot(x::AbstractVector)
	unique(x) .== permutedims(x)
end;

# ╔═╡ 85837bce-cde5-4971-9e28-e9530595246a
function prepMNIST(samples::Int)
	# x is a 28x28x60000 tensor
	# y is a 1x60000 tensor
	dt = MNIST(:train, dir="tmp/")
	x, y = dt[:]

	# then we want to flatten the 28x28 into a single dim for this assignment
	x = reshape(x, 28*28, size(x, 3))

	# and now we want to reorganize so we have a 60000x784
	x = x'

	# one hot encode the first dimension
	y = oneHot(y)
	# and make a 60000x10
	y = y'

	x[1:samples, :], y[1:samples, :]
end;

# ╔═╡ 405886d9-78fe-4fba-9080-6560c86eb113
toy_x, toy_y = prepMNIST(500)

# ╔═╡ 6f1472e6-8f0b-40f6-baa7-f6ab4dcb489d
# define a model to hold our learnable parameters
mutable struct SAE
	W2::Matrix
	Wx::Matrix
	Wy::Matrix
	function SAE(features, hidden, outputs, rng)
		new(
			rand(rng, Normal(0, 1/(features+hidden)), features, hidden),
			rand(rng, Normal(0, 1/(features+hidden)), hidden, features),
			rand(rng, Normal(0, 1/(outputs+hidden)), hidden, outputs),
		)
	end
end

# ╔═╡ b5237a83-3c00-4008-8f23-61fb90ca755d
# define a separate model to hold our tuning parameters
# (hyperparameters, meta-parameters, etc.)
struct SAEParams
	β::Float64
	α::Float64
	epochs::Int
	hidden::Int
	SAEParams(vals) = new(vals...)
	SAEParams(β, α, epochs, hidden) = new(β, α, epochs, hidden)
end

# ╔═╡ a59ccf48-3836-48b5-b28f-ab44d41637d5
md"""
## Utility functions
Here, we implement a few utility functions which we will need throughout the notebook.
"""

# ╔═╡ d6fabf73-d039-48b1-ba02-5de61339b4d2
"""
	softmax(X::AbstractMatrix)

Compute the softmax of each row of a matrix
"""
function softmax(X::AbstractMatrix)
	exp.(X) ./ (sum(exp.(X), dims=2))
end

# ╔═╡ 01a43fcd-a999-4533-9826-2d3c52f4d090
# takes a model and some data
# and generates the predictions for that model for both heads
function forward(sae::SAE, X::AbstractMatrix)
	# size is samples x hidden 
	h = X * sae.W2

	# size is samples x features 
	X̂ = h * sae.Wx

	# size is samples x outputs 
	z = h * sae.Wy
	Ŷ = softmax(z)

	X̂, Ŷ
end

# ╔═╡ 16564236-60a1-4d92-85f0-441531acd6be
# takes each row of a matrix and replaces it with a one-hot version of that row
# where the one is in the position of the maximizing value
function oneHotMax(X::AbstractMatrix)
	function _inner(a)
		idx = argmax(a)
		out = zeros(Int, size(a))
		out[idx] = 1

		return out
	end

	hcat(map(_inner, eachrow(X))...)'
end

# ╔═╡ 8c115842-4f93-4965-9ba5-9f752d0fa669
"""
	gradient(sae::SAE, params::SAEParams, X::AbstractMatrix, Y::AbstractMatrix)

Compute the gradient of the SAE parameters given inputs `(X, Y)`
"""
function gradient(sae::SAE, params::SAEParams, X::AbstractMatrix, Y::AbstractMatrix)
	# samples x hidden matrix
	h = X * sae.W2

	# samples x features matrix
	X̂ = h * sae.Wx

	# samples x outputs matrix
	z = h * sae.Wy
	Ŷ = softmax(z)

	# here is the gradient for the Wx head
	δ_wx = params.β * (X̂ - X) # (samples, features)
	∇Wx = h' * δ_wx           # (hidden, features)

	# TODO: compute the gradient for the Wy head
	# HINT: use your solution to Q1a in the written portion
	### BEGIN SOLUTION
	∇Wy = zeros(1, 1)

	### END SOLUTION

	# TODO: compute the gradient for W2
	### BEGIN SOLUTION
	∇W2 = zeros(1, 1)

	### END SOLUTION

	∇Wx, ∇Wy, ∇W2
end;

# ╔═╡ aa921231-4c0f-48b1-8390-40778de3121c
# ##############
# Test Block
# ##############
begin
	# let's first double check that you are getting the right shapes
	# we'll ignore the data within these matrices for now
	__shape_check = let
		sae = SAE(28*28, 128, 10, Xoshiro(1))
		params = SAEParams(0.5, 0.001, 30, 128)

		∇Wx, ∇Wy, ∇W2 = gradient(sae, params, toy_x, toy_y)

		size(∇Wx) == (128, 28*28) && size(∇Wy) == (128, 10) && size(∇W2) == (28*28, 128)
	end

	# now let's make a small fake dataset so the gradients are little
	# and see if your gradients exactly match ours
	function __do_grad_check()
		# if shapes are wrong, obviously the computed values are wrong
		# this lets us avoid some thrown exceptions
		if !__shape_check
			return false
		end

		rng = Xoshiro(367)

		sae = SAE(3, 4, 2, Xoshiro(2))
		params = SAEParams(2.0, 0.01, 30, 4)

		x = ones(5, 3)
		y = rand(rng, Bernoulli(0.5), 5, 2)

		∇Wx, ∇Wy, ∇W2 = gradient(sae, params, x, y)

		expect_∇Wx = [-0.959675480821841 -1.0681537505986198 -0.9932776352234473; -1.7156436922085638 -1.9095738936187843 -1.7757153783105837; 1.4951414008681339 1.6641468151764258 1.547492402022379; 5.433633659507029 6.047832094043896 5.623887345088495]

		expect_∇Wy = [0.0489201775125635 0.14528383392392707; 0.08745622415952288 0.25972872938048425; -0.07621595445391038 -0.22634721769747854; -0.27698355170404837 -0.8225896628256475]

		expect_∇W2 = [-2.027220025407422 -0.7343071937158687 2.2714506331470448 -3.4635294528831597; -2.027220025407422 -0.7343071937158687 2.2714506331470448 -3.4635294528831597; -2.027220025407422 -0.7343071937158687 2.2714506331470448 -3.4635294528831597]

		expect_∇Wx ≈ ∇Wx && expect_∇Wy ≈ ∇Wy && expect_∇W2 ≈ ∇W2
	end

	__grad_check = __do_grad_check()
	__gradient_check = __shape_check && __grad_check

	md"""
	This cell performs a number of tests on your gradient implemenation. We checked:
	  * Does your gradient have the correct shapes? $(_check_complete(__shape_check))
	  * And does it compute the correct values for a very small fake dataset? $(_check_complete(__grad_check))
	"""
end

# ╔═╡ 7742ccb2-c30e-4b96-a522-0c9883c35905
md"""
# Supervised Autoencoders $(_check_complete(__gradient_check))

In this part of the notebook, you will implement a supervised autoencoder with a single hidden layer.
You will use backpropagation in order to take gradients w.r.t. the weights for a joint loss function:
```math
\ell(\mathbf{x}, \mathbf{y}; \mathbf{W}^{(2)}, \mathbf{W}_x^{(1)}, \mathbf{W}_y^{(1)}) = \ell_y(\hat{\mathbf{y}},\mathbf{y}) + \beta \ell_x(\hat{\mathbf{x}}, \mathbf{x})
```
where $\beta \ge 0$ is a hyperparameter that determines with how much weight the autoencoder portion of the SAE influences the hidden dimension.

For this assignment, we will use identity activations (no activation, resulting in a linear function) for every layer except the heads corresponding to $\hat{\mathbf{y}}$. For this supervised part, because we are doing multinomial logistic regression, we use a softmax activation.

As in the assignment pdf, use a multinomial cross-entropy for the "$\hat{\mathbf{y}}$" head of the neural network and a mean squared loss for the "$\hat{\mathbf{x}}$" head.
"""

# ╔═╡ fbcb4525-b780-4072-9101-d4e1a3e1d8f6
md"""
## Gradient Computation $(_check_complete(__gradient_check))
In this section, we will implement the gradient of the SAE loss function. The loss function is defined at the beginning of the _Supervised Autoencoders_ section.
"""

# ╔═╡ 816a5d46-1c95-408e-b5a6-bb51bac39ab4
"""
	train!(X::AbstractMatrix, Y::AbstractMatrix, params::SAEParams, seed::Int)

Train an SAE with the given parameters on the dataset `(X, Y)`. The `seed` parameter controls the seed to the random number generator.
"""
function train!(X::AbstractMatrix, Y::AbstractMatrix, params::SAEParams, seed::Int)
	rng = Xoshiro(seed)
	samples, features = size(X)
	classes = size(Y, 2)

	# Build the SAE model with N hidden units
	sae = SAE(features, params.hidden, classes, rng)

	# TODO: train the model using SGD for params.epochs number of epochs
	# use only a single sample for each iteration
	### BEGIN SOLUTION

	### END SOLUTION
	return sae
end;

# ╔═╡ 7656e820-32b0-42c9-bfac-6b8b6f71ea2d
# ################
# Test Block
# ################
__test_training_loop = try
	seed = 1234
	rng = Random.Xoshiro(seed)
	n = 20
	d = 5
	m = 2
	X1 = rand(rng, Normal(0.0, 1.0), n÷2, d)
	X2 = rand(rng, Normal(5.0, 1.0), n÷2, d)	
	perm = randperm(rng, n)
	X = vcat(X1, X2)[perm, :]
	X = (X .- mean(X,dims=1))./std(X,dims=1)
	Y = vcat(ones(eltype(X), size(X1,1)), 2*ones(eltype(X), size(X2,1)))[perm]
	Y = oneHot(Y)'
	params = SAEParams(0.1, 0.1, 10, 10)
	sae = train!(X, Y, params, seed)
	
	t1 = isclose(sae.Wx[1:2, :][:], [-0.03758060050829525, 0.274632702767149, 0.04804355016209671, 0.19679587016980385, 0.09201080057634131, 0.1686692616271758, 0.15803789415851136, 0.2258495378621631, 0.07131836140831797, 0.13929083936085002], 1.e-6)
	t2 = isclose(sae.Wy[1:5, :][:], [0.07743702827086997, 0.7255006982603416, 0.16480520763532078, -0.22292937416738245, 0.06798112956781598, -0.10128085363900588, -0.8339833809108191, -0.1291333057903954, 0.26814749353302464, 0.040451479119764835], 1.e-6)
	t3 = isclose(sae.W2[2:3, 4:8][:], [-0.21518623027353057, -0.19134709416224316, 0.0574352617540672, -0.14549702813838197, -0.1688520451271272, -0.21754027290309436, 0.1768971124051647, 0.08544550172779417, 0.06067206071617156, -0.05980604475513782], 1.e-6)
	t1 & t2 & t3
catch
	false
end

# ╔═╡ ff194fde-5f61-4cb7-91c5-22a4adede59a
md"
## Training Loop $(_check_complete(__test_training_loop))

Now, implement the training loop for the SAE, where you do SGD for multiple epochs. Remember to randomly shuffle the data for each epoch. Here we simply use vanilla SGD, with 1 sample per update (a mini-batch of size 1) and use a constant stepsize (rather than RMSProp and Adam). For this model and dataset, this provides sufficiently good performance. Of course, you can always test for yourself the impact of using mini-batches and improved stepsizes. But for your submission please use vanilla SGD.
"

# ╔═╡ 02b09245-38f2-4385-b0f1-53cf4136f64e
# First step is computing the number of true positives, true negatives, false positives, and false negatives.
begin
	# True positive
	function tp(y::AbstractVector, ŷ::AbstractVector)
		# default: 1
		sum(ŷ[y .== 1])
	end

	# True negative
	function tn(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	# False positive
	function fp(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	# False negative
	function fn(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end
end;

# ╔═╡ ccfa2549-1b63-43fb-9b0d-e1f96b1543ea
# ##############
# Test Block
# ##############
begin
	__y = [1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1]
	__ŷ = [0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0]

	__check_tp = tp(__y, __ŷ) == 3
	__check_tn = tn(__y, __ŷ) == 4
	__check_fp = fp(__y, __ŷ) == 6
	__check_fn = fn(__y, __ŷ) == 8
end

# ╔═╡ 40b3ec56-c62f-42d7-88a1-ca3405a6e598
# Use your definitions of tp, tn, fp, and fn to implement precision and recall
begin
	function tpr(y::AbstractVector, ŷ::AbstractVector)
		# default: 1
		tp(y, ŷ) / (tp(y, ŷ) + fn(y, ŷ))
	end
	
	function fpr(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end
	
	function tnr(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	function fnr(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	# Precision
	function pr(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	# Recall
	function rc(y::AbstractVector, ŷ::AbstractVector)
		### BEGIN SOLUTION
		1

		### END SOLUTION
	end

	function accuracy(y::AbstractVector, ŷ::AbstractVector)
		# default: 1
		sum(y .== ŷ) / size(y, 1)
	end

	function error(y::AbstractVector, ŷ::AbstractVector)
		# default: 1
		1 - accuracy(y, ŷ)
	end
end;

# ╔═╡ eb49d2cf-e388-431e-a616-45c05893b39d
# ##############
# Test Block
# ##############
begin
	__check_tpr = tpr(__y, __ŷ) ≈ 0.2727272727272727
	__check_fnr = fnr(__y, __ŷ) ≈ 0.7272727272727273
	__check_tnr = tnr(__y, __ŷ) ≈ 0.4
	__check_fpr = fpr(__y, __ŷ) ≈ 0.6
	__check_pr = pr(__y, __ŷ) ≈ 0.3333333333333333
	__check_rc = rc(__y, __ŷ) ≈ 0.2727272727272727
	__check_acc = accuracy(__y, __ŷ) ≈ 0.3333333333333333
	__check_err = error(__y, __ŷ) ≈ 0.6666666666666667
	__check_all = __check_tpr && __check_fpr && __check_tnr && __check_fnr && __check_acc && __check_rc && __check_pr && __check_err
end

# ╔═╡ 45a2f25e-56e8-4312-a658-6981cdd35fa2
md"""
# Evaluating our model $(_check_complete(__check_tp && __check_tn && __check_fp && __check_fn && __check_all))

Before we can do anything else (tuning hyperparameters, reporting statistics about our model, etc.), we must first decide how our model will be evaluated. In class, we primarily focused on classification accuracy. But other metrics are often used for classification, that better reflect how well the model performed for different classes and if it is skewed towards predicting one class more than others. These metrics can provide a more nuanced picture of the model. 

To start, we will treat each class separately: for example, what is the true positive rate for the digit 2?

## True/False Positives/Negatives $(_check_complete((__check_tp && __check_fp && __check_tn && __check_fn)))

To begin, let's implement fuctions which count the true/false positives and negatives. A _true positive_ is a prediction that was predicted positive when it has a postive label (which we represent as a 1). A _false positive_ is a prediction that was predicted positive when is has a negative label. A _true negative_ is a prediction that was predicted negative when it has a negative label (which we represent as a 0). A _false negative_ is a prediction that was predicted negative when is has a positive label. We provide the code for true positives. It is your job to complete the rest.
"""

# ╔═╡ 17398962-9183-4fa4-8eab-d5d9245f3d00
md"
Great! You've implemented four functions to count the number of:

* True positives: $(_check_complete(__check_tp))
* True negatives: $(_check_complete(__check_tn))
* False positives: $(_check_complete(__check_fp))
* False negatives: $(_check_complete(__check_fn))

## Evaluation criteria $(_check_complete(__check_all))

Often, we'd like to consider the rates of true positives/negatives and false positives/negatives. Additionally, we often want to consider the accuracy of our predictions as well as the _precision_ and _recall_, which are defined as:

```math
\text{Precision: } \quad pr(y, \hat y) = \frac{TP}{TP + FP}
```
```math
\text{Recall: } \quad rc(y, \hat y) = \frac{TP}{TP + FN}
```

These definitions extend even to the multiclass setting. We can compute the precision and recall for each class. With precision, we get a sense of: when we classified an item as class j, was it actually class j? Precision is high when we are careful about assigning class j to an item. For example, imagine there are 100 samples that are class j. Even if we only label 10 of these as class j and do not incorrectly label any other items as class j, then our Precision = 1 which is maximal (here TP = 10 and FP = 0). Mentally, we can replace the term TP with Correctly-Labeled-as-Class-j and FP with Incorrectly-Labeled-as-Class-j. 

Recall asks: did we find all the items that were labeled as class j? Recall can be high when precision is low. For example, if we label all items in our dataset as class j, then we would have Recall = 1 (False Negatives = 0). We can think of FN as Incorrectly-Not-Labeled-as-Class-j or Should-Have-Been-Labeled-as-Class-j. If we label everything as class j, then there are no items that Should-Have-Been-Labeled-as-Class-j, since it was already labeled class j. Of course, this labelling will have bad Precision. Ideally, we want to get high Recall and high Precision. 

Reporting more metrics gives a multi-faceted view on your learned predictor. We will report all of them here, so you can see what they would look like. We have implemented some of the below functions for you: True Positive Rate, Accuracy and Error. It is your job to fill in the rest. Again, you will see checkmarks beside the metrics once they are implemented correctly.

* True positive rate: $(_check_complete(__check_tpr))
* False positive rate: $(_check_complete((__check_fpr)))
* True negative rate: $(_check_complete((__check_tnr)))
* False negative rate: $(_check_complete((__check_fnr)))
* Precision: $(_check_complete((__check_pr)))
* Recall: $(_check_complete((__check_rc)))
* Accuracy: $(_check_complete((__check_acc)))
* Error: $(_check_complete((__check_err)))
"

# ╔═╡ 449545d1-9bb7-4a41-b2dc-66ef25d51417
let
	params = SAEParams(0.5, 0.01, 5, 16)
	sae = train!(toy_x, toy_y, params, 0)

	_, Ŷ = forward(sae, toy_x)

	# take our prediction matrix and replace the probabilities
	# with a one-hot encoding for the maximizing prediction
	Ŷ = oneHotMax(Ŷ)

	# 12 different metrics
	# 10 classes
	col_tp = zeros(Int, 10)
	col_tn = zeros(Int,10)
	col_fp = zeros(Int,10)
	col_fn = zeros(Int,10)
	col_tpr = zeros(10)
	col_fnr = zeros(10)
	col_tnr = zeros(10)
	col_fpr = zeros(10)
	col_pr = zeros(10)
	col_rc = zeros(10)
	col_acc = zeros(10)
	col_err = zeros(10)

	labels = ["Zero","One","Two","Three","Four","Five","Six","Seven","Eight","Nine"]
	
	for class in 1:10
		col_tp[class] = tp(toy_y[:, class], Ŷ[:, class])
		col_tn[class] = tn(toy_y[:, class], Ŷ[:, class])
		col_fp[class] = fp(toy_y[:, class], Ŷ[:, class])
		col_fn[class] = fn(toy_y[:, class], Ŷ[:, class])
		col_tpr[class] = tpr(toy_y[:, class], Ŷ[:, class])
		col_fnr[class] = fnr(toy_y[:, class], Ŷ[:, class])
		col_tnr[class] = tnr(toy_y[:, class], Ŷ[:, class])
		col_fpr[class] = fpr(toy_y[:, class], Ŷ[:, class])
		col_pr[class] = pr(toy_y[:, class], Ŷ[:, class])
		col_rc[class] = rc(toy_y[:, class], Ŷ[:, class])
		col_acc[class] = accuracy(toy_y[:, class], Ŷ[:, class])
		col_err[class] = error(toy_y[:, class], Ŷ[:, class])
	end

	DataFrame("Digit"=>labels, 
			  "True Positive"=>col_tp,
			  "True Negative"=>col_tn,
			  "False Positive"=>col_fp,
			  "False Negative"=>col_fn,		
			  "True Positive Rate"=>col_tpr,
			  "False Negative Rate"=>col_fnr,
			  "True Negative Rate"=>col_tnr,
			  "False Positive Rate"=>col_fpr,
			  "Precision"=>col_pr,
			  "Recall"=>col_rc,
			  "Accuracy"=>col_acc,
			  "Error"=>col_err)
end

# ╔═╡ 1e6f182f-af6e-4aba-a2db-e9d0f982d6e1
# A helper function which takes lists of beta, alpha, epoch, and hidden
# then turns that into a list of all combinations of SAEParams

function generateParamSettings(βs::Vector, αs::Vector, epochss::Vector{Int}, hiddens::Vector{Int})

	out = Iterators.product(βs, αs, epochss, hiddens)
	vcat(map(SAEParams, out)...)
end;

# ╔═╡ a57877b7-529b-477b-b904-d9d77b65397b
# Here are a few helper functions which you might want while implementing cross validation
begin
	# stacks matrices vertically. So if given matrices of shape
	# A=(10, 23), B=(15, 23), C=(2, 23)
	# then D=stack([A, B, C]) is a (27, 23) matrix
	function stack(arr)
		vcat(arr...)
	end

	# returns all elements of an array except the specified index
	function exclude(arr::Array, idx::Int)
		arr[1:end .!= idx]
	end
end;

# ╔═╡ 1a0e23c3-859b-41e4-bdb6-f4db4e4cef44
md"""
We can choose to optimize any of the above metrics that we would like. By default, we have selected accuracy. You can try other metrics, if you would like.
Below, we use `argmax` when selecting hyperparameters, because we want high accuracy (if we had errors, we would use `argmin`).
"""

# ╔═╡ b25dec91-e855-4537-aa49-f43a2fe34e19
# define the metric of interest for our model on some data
# here we will take the average accuracy over the classes
# however, we could use any of the above metrics here
function metric(sae::SAE, X::AbstractMatrix, Y::AbstractMatrix; f = accuracy)
	_, Ŷ = forward(sae, X)

	Ŷ = oneHotMax(Ŷ)

	classes = size(Y, 2)

	metrics = zeros(classes)
	for class in 1:classes
		metrics[class] = f(Y[:, class], Ŷ[:, class])
	end

	return mean(metrics)
end;

# ╔═╡ 4fabdb20-bac5-4193-a0d4-73195346ef37
md"""
Now you will implement `internal_cross_validation`.
Make sure that each of your `k` folds are non-overlapping and that every validation set is used exactly once.
We explicitly fix the random seed when training the model.
We want to reduce the number of confounding factors while selecting hyperparameters, for instance starting each neural network with the exact same set of weights. That way we are removing one potential source of stochasticity that could give noisier estimates of the performance for a hyperparameter.
"""

# ╔═╡ 393e0f9e-81a0-4f5e-b95c-d63dffbc4deb
function internal_cross_validation(settings::Array{SAEParams}, k::Int, X::AbstractMatrix, Y::AbstractMatrix; comparator=accuracy)
	# TODO: first you will need a consistent set of train/validate splits so that each hyperparameter setting sees the exact same sets of data. The goal here is to partition the data into k folds and then create the k pairs of train/validate datasets
	# Hint: there is no need to shuffle the data when splitting in this first part
	Xs = []
	Ys = []

	### BEGIN SOLUTION

	### END SOLUTION

	# Warning! Do not change this! Every param setting and fold should see the *exact* same random seed
	seed = 0

	# TODO: then for every parameter setting, we need to know the average accuracy of the model for that setting on the holdout set

	average_metrics = zeros(length(settings))

	### BEGIN SOLUTION

	### END SOLUTION

	best_idx = argmax(average_metrics)
	best_params = settings[best_idx]
	return best_params
end;

# ╔═╡ bb9c5701-15ff-48bb-b4dc-1643fe758ab3
# ##############
# Test Block
# ##############
begin
	__check_internal_cross_val = let
		rng = Xoshiro(1)
		Random.seed!(367) 
		settings = generateParamSettings([100, 1., 0.1], [0.001, 0.1], [1, 3, 5], [2, 4, 6])

		__x = rand(rng, Normal(0, 3), 100, 3)
		__y = rand(rng, Bernoulli(0.5), 100, 2)  

		got = internal_cross_validation(settings, 3, __x, __y, comparator=accuracy)

		got == SAEParams(0.1, 0.1, 1, 2)
	end
end

# ╔═╡ 36467250-5f49-4149-8856-67b68419c0ec
md"""
# Selecting Hyperparameters $(_check_complete(__check_internal_cross_val))

The supervised autoencoder class has 4 hyperparameters that need to be selected (for our implementation).
So far we have worked with toy data and a few hardcoded hyperparameters that we know work okay, but now it's time to start tuning our algorithm to get the best performance.

## Internal K-Fold Cross-Validation
In this section, you will implement **internal k-fold cross-validation** to select hyperparameters.
"""

# ╔═╡ d75cf0fd-fd8e-45f7-8d32-b1df812aa407
md"""
If all went well with your cross-validation code, then it will pass our tests and you should see a happy check mark right beside this sentence. $(_check_complete(__check_internal_cross_val)).

Now, let's run internal CV on our small MNIST dataset, to see which hyperparameter settings are picked.
"""

# ╔═╡ a7a9a2d8-7172-4cb9-9d8a-0ccf4dfe3190
# Note this cell will take several seconds to complete (i.e. ~30s)
best_toy_param = let
	settings = generateParamSettings(
		[1.0, 0.1],    # betas
		[0.01, 0.001], # alphas
		[2],           # epochs
		[16, 32],      # hidden units
	)

	# of these parameter settings, let's see which one is the best on our small 
	# dataset note how long this takes to run! This is only for a few hundred 
	# samples, imagine running this for bigger models and datasets
	# feel free to change the `comparator` parameter to try other metrics
	internal_cross_validation(settings, 4, toy_x, toy_y, comparator=accuracy)
end

# ╔═╡ 8c0fbee2-1cde-460f-b914-c6e9f7781fc4
function repeated_random_sampling(params::SAEParams, X::AbstractMatrix, Y::AbstractMatrix, k::Int, train_percent::Float64, rng::AbstractRNG; comparator=accuracy)

	samples = size(X, 1)
	train_samples = floor(Int, samples * train_percent)

	# We still want to ensure that the models are initialized in the same way across partitions, because the learned models will be more representative of the actual model we deploy, which was trained with seed = 0
	seed = 0
	
	metrics = zeros(k)
	### BEGIN SOLUTION

	### END SOLUTION

	mean(metrics), std(metrics) / sqrt(k)
end

# ╔═╡ 56807acf-7530-4fdd-bca3-9ff7cc06ba6e
# ##############
# Test Block
# ##############
begin
	__check_rrs = let
		params = SAEParams(0.1, 0.001, 2, 10)
		rng = Xoshiro(1)

		X = rand(rng, Normal(1, 20), 500, 30)
		Y = rand(rng, Bernoulli(0.4), 500, 3)

		acc, stderr = repeated_random_sampling(params, X, Y, 9, 0.42, rng, comparator=accuracy)

		acc >= 0.5 && stderr <= 0.01
	end
end

# ╔═╡ 8f157646-95b2-4a0f-8497-a51378886b83
md"""
# Evaluating Generalization Error $(_check_complete(__check_rrs))

Internal CV was used above to select hyperparameters best-hypers for the SAE. The model f learned with SAE+best-hypers on the entire dataset is the final model that we want to deploy. However, before deploying, we want to get a sense of its generalization performance (classification accuracy in deployment). In the notes we discussed using external CV, where we would split up the dataset into train/validate pairs like above, and then call SAE+internal CV on these train/validate pairs. This is called nested cross-validation: we have nested for loops where in the outer loop we iterate over the splits from external CV, and inside when we call SAE+internal CV, it splits the data further into multiple folds and loops over all hypers and those partitionings. 

Unfortunately, this is very expensive. And already running on this MNIST dataset can take a bit too much compute for a student laptop. So, we are going to do a standard approximation used in practice (and keep in mind that we cheated a little bit and introduced more bias). Instead of using nest cross-validation to properly evaluate f, we will instead just call cross-validation on SAE+best-hypers. To implement cross-validation to evaluate SAE+best-hypers, we will use repeated random sampling (RRS). The main reason we use RRS instead of k-fold for our external CV is because it actually helps reduce bias a little bit from this un-nested procedure, and because it lets you implement the other way approach we discussed to generated partitions for CV. We will report the average and standard error of our accuracy across these partitions.

A few notes:
 * The train/test sets must be non-overlapping
 * Each train/test split needs to be sampled randomly `k` times
 * You should use `train_percent` percentage of the samples in the train set, and the remainder in the test set (i.e. if `train_percent=0.9`, then 90% of the samples are training samples and 10% are testing samples)
"""

# ╔═╡ cafe4f1a-8532-4528-9ab7-f6450835d879
md"""
Does RRS report the correct accuracy and standard error on a small toy dataset? $(_check_complete(__check_rrs))
"""

# ╔═╡ 5bbbc793-61ff-430b-a74b-3f1207e4ebbc
let
	repeated_random_sampling(best_toy_param, toy_x, toy_y, 10, .90, Xoshiro(3))
end;

# ╔═╡ 909e0c4b-4741-4337-91af-b663c55daee9
md"""
# Putting it all together

Finally, we will run the SAE with internal CV to get best-hypers on the full MNIST dataset and then run external CV to evaluate the model learned with those hypers. If you want to examine other metrics, you can change the metric used in RRS and the outputted numbers will be the mean and standard error across the folds for that metric. 
"""

# ╔═╡ 5df77b3e-91a4-4965-9711-58a9473b3c29
function runExperiment()
	X, Y = prepMNIST(50000)

	settings = generateParamSettings(
		[0.1, 0.],     # beta
		[0.001, 0.01], # alpha
		[2],           # epochs
		[100],         # hidden units
	)

	best_params = internal_cross_validation(settings, 3, X, Y)

	best_params

	acc, stderr = repeated_random_sampling(best_params, X, Y, 10, 0.75, Xoshiro(734))
end;

# ╔═╡ 2b585960-55f5-4b65-b942-9b5653c39a6c
## uncomment the following when you're ready to run it
## it will take some time, ~20m on a modern laptop
## if it is taking too long to run, you can simply skip this cell
## and answer the corresponding questions by inspecting the above code

runExperiment()

# ╔═╡ 00000000-0000-0000-0000-000000000001
PLUTO_PROJECT_TOML_CONTENTS = """
[deps]
CSV = "336ed68f-0bac-5ca0-87d4-7b16caf5d00b"
DataFrames = "a93c6f00-e57d-5684-b7b6-d8193f3e46c0"
Distributions = "31c24e10-a181-5473-b8eb-7969acd0382f"
LinearAlgebra = "37e2e46d-f89d-539d-b4ee-838fcccc9c8e"
MLDatasets = "eb30cadb-4394-5ae3-aed4-317e484a6458"
PlutoUI = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
Random = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"
Statistics = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"
StatsBase = "2913bbd2-ae8a-5f71-8c99-4fb6c76f3a91"

[compat]
CSV = "~0.10.4"
DataFrames = "~1.3.6"
Distributions = "~0.25.76"
MLDatasets = "~0.7.5"
PlutoUI = "~0.7.44"
StatsBase = "~0.33.21"
"""

# ╔═╡ 00000000-0000-0000-0000-000000000002
PLUTO_MANIFEST_TOML_CONTENTS = """
# This file is machine-generated - editing it directly is not advised

julia_version = "1.8.0"
manifest_format = "2.0"
project_hash = "dae622f11c5e9bdc010cf7b487851407bacf1e7a"

[[deps.AbstractFFTs]]
deps = ["ChainRulesCore", "LinearAlgebra"]
git-tree-sha1 = "69f7020bd72f069c219b5e8c236c1fa90d2cb409"
uuid = "621f4979-c628-5d54-868e-fcf4e3e8185c"
version = "1.2.1"

[[deps.AbstractPlutoDingetjes]]
deps = ["Pkg"]
git-tree-sha1 = "8eaf9f1b4921132a4cff3f36a1d9ba923b14a481"
uuid = "6e696c72-6542-2067-7265-42206c756150"
version = "1.1.4"

[[deps.Accessors]]
deps = ["Compat", "CompositionsBase", "ConstructionBase", "Dates", "InverseFunctions", "LinearAlgebra", "MacroTools", "Requires", "Test"]
git-tree-sha1 = "ce67f55da3a937bb001a8d00559bdfa4dba6e4f5"
uuid = "7d9f7c33-5ae7-4f3b-8dc6-eff91059b697"
version = "0.1.20"

[[deps.Adapt]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "195c5505521008abea5aee4f96930717958eac6f"
uuid = "79e6a3ab-5dfb-504d-930d-738a2a938a0e"
version = "3.4.0"

[[deps.ArgCheck]]
git-tree-sha1 = "a3a402a35a2f7e0b87828ccabbd5ebfbebe356b4"
uuid = "dce04be8-c92d-5529-be00-80e4d2c0e197"
version = "2.3.0"

[[deps.ArgTools]]
uuid = "0dad84c5-d112-42e6-8d28-ef12dabb789f"
version = "1.1.1"

[[deps.Artifacts]]
uuid = "56f22d72-fd6d-98f1-02f0-08ddc0907c33"

[[deps.BangBang]]
deps = ["Compat", "ConstructionBase", "Future", "InitialValues", "LinearAlgebra", "Requires", "Setfield", "Tables", "ZygoteRules"]
git-tree-sha1 = "7fe6d92c4f281cf4ca6f2fba0ce7b299742da7ca"
uuid = "198e06fe-97b7-11e9-32a5-e1d131e6ad66"
version = "0.3.37"

[[deps.Base64]]
uuid = "2a0f44e3-6c83-55bd-87e4-b1978d98bd5f"

[[deps.Baselet]]
git-tree-sha1 = "aebf55e6d7795e02ca500a689d326ac979aaf89e"
uuid = "9718e550-a3fa-408a-8086-8db961cd8217"
version = "0.1.1"

[[deps.BitFlags]]
git-tree-sha1 = "84259bb6172806304b9101094a7cc4bc6f56dbc6"
uuid = "d1d4a3ce-64b1-5f1a-9ba4-7e7e69966f35"
version = "0.1.5"

[[deps.BufferedStreams]]
git-tree-sha1 = "bb065b14d7f941b8617bc323063dbe79f55d16ea"
uuid = "e1450e63-4bb3-523b-b2a4-4ffa8c0fd77d"
version = "1.1.0"

[[deps.CSV]]
deps = ["CodecZlib", "Dates", "FilePathsBase", "InlineStrings", "Mmap", "Parsers", "PooledArrays", "SentinelArrays", "Tables", "Unicode", "WeakRefStrings"]
git-tree-sha1 = "873fb188a4b9d76549b81465b1f75c82aaf59238"
uuid = "336ed68f-0bac-5ca0-87d4-7b16caf5d00b"
version = "0.10.4"

[[deps.Calculus]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "f641eb0a4f00c343bbc32346e1217b86f3ce9dad"
uuid = "49dc2e85-a5d0-5ad3-a950-438e2897f1b9"
version = "0.5.1"

[[deps.ChainRulesCore]]
deps = ["Compat", "LinearAlgebra", "SparseArrays"]
git-tree-sha1 = "e7ff6cadf743c098e08fca25c91103ee4303c9bb"
uuid = "d360d2e6-b24c-11e9-a2a3-2a2ae2dbcce4"
version = "1.15.6"

[[deps.ChangesOfVariables]]
deps = ["ChainRulesCore", "LinearAlgebra", "Test"]
git-tree-sha1 = "38f7a08f19d8810338d4f5085211c7dfa5d5bdd8"
uuid = "9e997f8a-9a97-42d5-a9f1-ce6bfc15e2c0"
version = "0.1.4"

[[deps.CodecZlib]]
deps = ["TranscodingStreams", "Zlib_jll"]
git-tree-sha1 = "ded953804d019afa9a3f98981d99b33e3db7b6da"
uuid = "944b1d66-785c-5afd-91f1-9de20f533193"
version = "0.7.0"

[[deps.ColorTypes]]
deps = ["FixedPointNumbers", "Random"]
git-tree-sha1 = "eb7f0f8307f71fac7c606984ea5fb2817275d6e4"
uuid = "3da002f7-5984-5a60-b8a6-cbb66c0b333f"
version = "0.11.4"

[[deps.ColorVectorSpace]]
deps = ["ColorTypes", "FixedPointNumbers", "LinearAlgebra", "SpecialFunctions", "Statistics", "TensorCore"]
git-tree-sha1 = "d08c20eef1f2cbc6e60fd3612ac4340b89fea322"
uuid = "c3611d14-8923-5661-9e6a-0046d554d3a4"
version = "0.9.9"

[[deps.Colors]]
deps = ["ColorTypes", "FixedPointNumbers", "Reexport"]
git-tree-sha1 = "417b0ed7b8b838aa6ca0a87aadf1bb9eb111ce40"
uuid = "5ae59095-9a9b-59fe-a467-6f913c188581"
version = "0.12.8"

[[deps.Compat]]
deps = ["Base64", "Dates", "DelimitedFiles", "Distributed", "InteractiveUtils", "LibGit2", "Libdl", "LinearAlgebra", "Markdown", "Mmap", "Pkg", "Printf", "REPL", "Random", "SHA", "Serialization", "SharedArrays", "Sockets", "SparseArrays", "Statistics", "Test", "UUIDs", "Unicode"]
git-tree-sha1 = "78bee250c6826e1cf805a88b7f1e86025275d208"
uuid = "34da2185-b29b-5c13-b0c7-acf172513d20"
version = "3.46.0"

[[deps.CompilerSupportLibraries_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "e66e0078-7015-5450-92f7-15fbd957f2ae"
version = "0.5.2+0"

[[deps.CompositionsBase]]
git-tree-sha1 = "455419f7e328a1a2493cabc6428d79e951349769"
uuid = "a33af91c-f02d-484b-be07-31d278c5ca2b"
version = "0.1.1"

[[deps.ConstructionBase]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "fb21ddd70a051d882a1686a5a550990bbe371a95"
uuid = "187b0558-2788-49d3-abe0-74a17ed4e7c9"
version = "1.4.1"

[[deps.ContextVariablesX]]
deps = ["Compat", "Logging", "UUIDs"]
git-tree-sha1 = "25cc3803f1030ab855e383129dcd3dc294e322cc"
uuid = "6add18c4-b38d-439d-96f6-d6bc489c04c5"
version = "0.1.3"

[[deps.Crayons]]
git-tree-sha1 = "249fe38abf76d48563e2f4556bebd215aa317e15"
uuid = "a8cc5b0e-0ffa-5ad4-8c14-923d3ee1735f"
version = "4.1.1"

[[deps.DataAPI]]
git-tree-sha1 = "46d2680e618f8abd007bce0c3026cb0c4a8f2032"
uuid = "9a962f9c-6df0-11e9-0e5d-c546b8b5ee8a"
version = "1.12.0"

[[deps.DataDeps]]
deps = ["HTTP", "Libdl", "Reexport", "SHA", "p7zip_jll"]
git-tree-sha1 = "bc0a264d3e7b3eeb0b6fc9f6481f970697f29805"
uuid = "124859b0-ceae-595e-8997-d05f6a7a8dfe"
version = "0.7.10"

[[deps.DataFrames]]
deps = ["Compat", "DataAPI", "Future", "InvertedIndices", "IteratorInterfaceExtensions", "LinearAlgebra", "Markdown", "Missings", "PooledArrays", "PrettyTables", "Printf", "REPL", "Reexport", "SortingAlgorithms", "Statistics", "TableTraits", "Tables", "Unicode"]
git-tree-sha1 = "db2a9cb664fcea7836da4b414c3278d71dd602d2"
uuid = "a93c6f00-e57d-5684-b7b6-d8193f3e46c0"
version = "1.3.6"

[[deps.DataStructures]]
deps = ["Compat", "InteractiveUtils", "OrderedCollections"]
git-tree-sha1 = "d1fff3a548102f48987a52a2e0d114fa97d730f0"
uuid = "864edb3b-99cc-5e75-8d2d-829cb0a9cfe8"
version = "0.18.13"

[[deps.DataValueInterfaces]]
git-tree-sha1 = "bfc1187b79289637fa0ef6d4436ebdfe6905cbd6"
uuid = "e2d170a0-9d28-54be-80f0-106bbe20a464"
version = "1.0.0"

[[deps.Dates]]
deps = ["Printf"]
uuid = "ade2ca70-3891-5945-98fb-dc099432e06a"

[[deps.DefineSingletons]]
git-tree-sha1 = "0fba8b706d0178b4dc7fd44a96a92382c9065c2c"
uuid = "244e2a9f-e319-4986-a169-4d1fe445cd52"
version = "0.1.2"

[[deps.DelimitedFiles]]
deps = ["Mmap"]
uuid = "8bb1440f-4735-579b-a4ab-409b98df4dab"

[[deps.DensityInterface]]
deps = ["InverseFunctions", "Test"]
git-tree-sha1 = "80c3e8639e3353e5d2912fb3a1916b8455e2494b"
uuid = "b429d917-457f-4dbc-8f4c-0cc954292b1d"
version = "0.4.0"

[[deps.Distributed]]
deps = ["Random", "Serialization", "Sockets"]
uuid = "8ba89e20-285c-5b6f-9357-94700520ee1b"

[[deps.Distributions]]
deps = ["ChainRulesCore", "DensityInterface", "FillArrays", "LinearAlgebra", "PDMats", "Printf", "QuadGK", "Random", "SparseArrays", "SpecialFunctions", "Statistics", "StatsBase", "StatsFuns", "Test"]
git-tree-sha1 = "04db820ebcfc1e053bd8cbb8d8bccf0ff3ead3f7"
uuid = "31c24e10-a181-5473-b8eb-7969acd0382f"
version = "0.25.76"

[[deps.DocStringExtensions]]
deps = ["LibGit2"]
git-tree-sha1 = "5158c2b41018c5f7eb1470d558127ac274eca0c9"
uuid = "ffbed154-4ef7-542d-bbb7-c09d3a79fcae"
version = "0.9.1"

[[deps.Downloads]]
deps = ["ArgTools", "FileWatching", "LibCURL", "NetworkOptions"]
uuid = "f43a241f-c20a-4ad4-852c-f6b1247861c6"
version = "1.6.0"

[[deps.DualNumbers]]
deps = ["Calculus", "NaNMath", "SpecialFunctions"]
git-tree-sha1 = "5837a837389fccf076445fce071c8ddaea35a566"
uuid = "fa6b7ba4-c1ee-5f82-b5fc-ecf0adba8f74"
version = "0.6.8"

[[deps.FLoops]]
deps = ["BangBang", "Compat", "FLoopsBase", "InitialValues", "JuliaVariables", "MLStyle", "Serialization", "Setfield", "Transducers"]
git-tree-sha1 = "ffb97765602e3cbe59a0589d237bf07f245a8576"
uuid = "cc61a311-1640-44b5-9fba-1b764f453329"
version = "0.2.1"

[[deps.FLoopsBase]]
deps = ["ContextVariablesX"]
git-tree-sha1 = "656f7a6859be8673bf1f35da5670246b923964f7"
uuid = "b9860ae5-e623-471e-878b-f6a53c775ea6"
version = "0.1.1"

[[deps.FileIO]]
deps = ["Pkg", "Requires", "UUIDs"]
git-tree-sha1 = "7be5f99f7d15578798f338f5433b6c432ea8037b"
uuid = "5789e2e9-d7fb-5bc7-8068-2c6fae9b9549"
version = "1.16.0"

[[deps.FilePathsBase]]
deps = ["Compat", "Dates", "Mmap", "Printf", "Test", "UUIDs"]
git-tree-sha1 = "e27c4ebe80e8699540f2d6c805cc12203b614f12"
uuid = "48062228-2e41-5def-b9a4-89aafe57970f"
version = "0.9.20"

[[deps.FileWatching]]
uuid = "7b1f6079-737a-58dc-b8bc-7a2ca5c1b5ee"

[[deps.FillArrays]]
deps = ["LinearAlgebra", "Random", "SparseArrays", "Statistics"]
git-tree-sha1 = "802bfc139833d2ba893dd9e62ba1767c88d708ae"
uuid = "1a297f60-69ca-5386-bcde-b61e274b549b"
version = "0.13.5"

[[deps.FixedPointNumbers]]
deps = ["Statistics"]
git-tree-sha1 = "335bfdceacc84c5cdf16aadc768aa5ddfc5383cc"
uuid = "53c48c17-4a7d-5ca2-90c5-79b7896eea93"
version = "0.8.4"

[[deps.FoldsThreads]]
deps = ["Accessors", "FunctionWrappers", "InitialValues", "SplittablesBase", "Transducers"]
git-tree-sha1 = "eb8e1989b9028f7e0985b4268dabe94682249025"
uuid = "9c68100b-dfe1-47cf-94c8-95104e173443"
version = "0.1.1"

[[deps.Formatting]]
deps = ["Printf"]
git-tree-sha1 = "8339d61043228fdd3eb658d86c926cb282ae72a8"
uuid = "59287772-0a20-5a39-b81b-1366585eb4c0"
version = "0.4.2"

[[deps.FunctionWrappers]]
git-tree-sha1 = "d62485945ce5ae9c0c48f124a84998d755bae00e"
uuid = "069b7b12-0de2-55c6-9aab-29f3d0a68a2e"
version = "1.1.3"

[[deps.Future]]
deps = ["Random"]
uuid = "9fa8497b-333b-5362-9e8d-4d0656e87820"

[[deps.GZip]]
deps = ["Libdl"]
git-tree-sha1 = "039be665faf0b8ae36e089cd694233f5dee3f7d6"
uuid = "92fee26a-97fe-5a0c-ad85-20a5f3185b63"
version = "0.5.1"

[[deps.Glob]]
git-tree-sha1 = "4df9f7e06108728ebf00a0a11edee4b29a482bb2"
uuid = "c27321d9-0574-5035-807b-f59d2c89b15c"
version = "1.3.0"

[[deps.Graphics]]
deps = ["Colors", "LinearAlgebra", "NaNMath"]
git-tree-sha1 = "d61890399bc535850c4bf08e4e0d3a7ad0f21cbd"
uuid = "a2bd30eb-e257-5431-a919-1863eab51364"
version = "1.1.2"

[[deps.HDF5]]
deps = ["Compat", "HDF5_jll", "Libdl", "Mmap", "Random", "Requires"]
git-tree-sha1 = "19effd6b5af759c8aaeb9c77f89422d3f975ab65"
uuid = "f67ccb44-e63f-5c2f-98bd-6dc0ccc4ba2f"
version = "0.16.12"

[[deps.HDF5_jll]]
deps = ["Artifacts", "JLLWrappers", "LibCURL_jll", "Libdl", "OpenSSL_jll", "Pkg", "Zlib_jll"]
git-tree-sha1 = "4cc2bb72df6ff40b055295fdef6d92955f9dede8"
uuid = "0234f1f7-429e-5d53-9886-15a909be8d59"
version = "1.12.2+2"

[[deps.HTTP]]
deps = ["Base64", "CodecZlib", "Dates", "IniFile", "Logging", "LoggingExtras", "MbedTLS", "NetworkOptions", "OpenSSL", "Random", "SimpleBufferStream", "Sockets", "URIs", "UUIDs"]
git-tree-sha1 = "3cdd8948c55d8b53b5323f23c9581555dc2e30e1"
uuid = "cd3eb016-35fb-5094-929b-558a96fad6f3"
version = "1.5.0"

[[deps.HypergeometricFunctions]]
deps = ["DualNumbers", "LinearAlgebra", "OpenLibm_jll", "SpecialFunctions", "Test"]
git-tree-sha1 = "709d864e3ed6e3545230601f94e11ebc65994641"
uuid = "34004b35-14d8-5ef3-9330-4cdb6864b03a"
version = "0.3.11"

[[deps.Hyperscript]]
deps = ["Test"]
git-tree-sha1 = "8d511d5b81240fc8e6802386302675bdf47737b9"
uuid = "47d2ed2b-36de-50cf-bf87-49c2cf4b8b91"
version = "0.0.4"

[[deps.HypertextLiteral]]
deps = ["Tricks"]
git-tree-sha1 = "c47c5fa4c5308f27ccaac35504858d8914e102f9"
uuid = "ac1192a8-f4b3-4bfe-ba22-af5b92cd3ab2"
version = "0.9.4"

[[deps.IOCapture]]
deps = ["Logging", "Random"]
git-tree-sha1 = "f7be53659ab06ddc986428d3a9dcc95f6fa6705a"
uuid = "b5f81e59-6552-4d32-b1f0-c071b021bf89"
version = "0.2.2"

[[deps.ImageBase]]
deps = ["ImageCore", "Reexport"]
git-tree-sha1 = "b51bb8cae22c66d0f6357e3bcb6363145ef20835"
uuid = "c817782e-172a-44cc-b673-b171935fbb9e"
version = "0.1.5"

[[deps.ImageCore]]
deps = ["AbstractFFTs", "ColorVectorSpace", "Colors", "FixedPointNumbers", "Graphics", "MappedArrays", "MosaicViews", "OffsetArrays", "PaddedViews", "Reexport"]
git-tree-sha1 = "acf614720ef026d38400b3817614c45882d75500"
uuid = "a09fc81d-aa75-5fe9-8630-4744c3626534"
version = "0.9.4"

[[deps.ImageShow]]
deps = ["Base64", "FileIO", "ImageBase", "ImageCore", "OffsetArrays", "StackViews"]
git-tree-sha1 = "b563cf9ae75a635592fc73d3eb78b86220e55bd8"
uuid = "4e3cecfd-b093-5904-9786-8bbb286a6a31"
version = "0.3.6"

[[deps.IniFile]]
git-tree-sha1 = "f550e6e32074c939295eb5ea6de31849ac2c9625"
uuid = "83e8ac13-25f8-5344-8a64-a9f2b223428f"
version = "0.5.1"

[[deps.InitialValues]]
git-tree-sha1 = "4da0f88e9a39111c2fa3add390ab15f3a44f3ca3"
uuid = "22cec73e-a1b8-11e9-2c92-598750a2cf9c"
version = "0.3.1"

[[deps.InlineStrings]]
deps = ["Parsers"]
git-tree-sha1 = "db619c421554e1e7e07491b85a8f4b96b3f04ca0"
uuid = "842dd82b-1e85-43dc-bf29-5d0ee9dffc48"
version = "1.2.2"

[[deps.InteractiveUtils]]
deps = ["Markdown"]
uuid = "b77e0a4c-d291-57a0-90e8-8db25a27a240"

[[deps.InternedStrings]]
deps = ["Random", "Test"]
git-tree-sha1 = "eb05b5625bc5d821b8075a77e4c421933e20c76b"
uuid = "7d512f48-7fb1-5a58-b986-67e6dc259f01"
version = "0.7.0"

[[deps.InverseFunctions]]
deps = ["Test"]
git-tree-sha1 = "49510dfcb407e572524ba94aeae2fced1f3feb0f"
uuid = "3587e190-3f89-42d0-90ee-14403ec27112"
version = "0.1.8"

[[deps.InvertedIndices]]
git-tree-sha1 = "bee5f1ef5bf65df56bdd2e40447590b272a5471f"
uuid = "41ab1584-1d38-5bbf-9106-f11c6c58b48f"
version = "1.1.0"

[[deps.IrrationalConstants]]
git-tree-sha1 = "7fd44fd4ff43fc60815f8e764c0f352b83c49151"
uuid = "92d709cd-6900-40b7-9082-c6be49f344b6"
version = "0.1.1"

[[deps.IteratorInterfaceExtensions]]
git-tree-sha1 = "a3f24677c21f5bbe9d2a714f95dcd58337fb2856"
uuid = "82899510-4779-5014-852e-03e436cf321d"
version = "1.0.0"

[[deps.JLD2]]
deps = ["FileIO", "MacroTools", "Mmap", "OrderedCollections", "Pkg", "Printf", "Reexport", "TranscodingStreams", "UUIDs"]
git-tree-sha1 = "1c3ff7416cb727ebf4bab0491a56a296d7b8cf1d"
uuid = "033835bb-8acc-5ee8-8aae-3f567f8a3819"
version = "0.4.25"

[[deps.JLLWrappers]]
deps = ["Preferences"]
git-tree-sha1 = "abc9885a7ca2052a736a600f7fa66209f96506e1"
uuid = "692b3bcd-3c85-4b1f-b108-f13ce0eb3210"
version = "1.4.1"

[[deps.JSON]]
deps = ["Dates", "Mmap", "Parsers", "Unicode"]
git-tree-sha1 = "3c837543ddb02250ef42f4738347454f95079d4e"
uuid = "682c06a0-de6a-54ab-a142-c8b1cf79cde6"
version = "0.21.3"

[[deps.JSON3]]
deps = ["Dates", "Mmap", "Parsers", "StructTypes", "UUIDs"]
git-tree-sha1 = "f1572de22c866dc92aea032bc89c2b137cbddd6a"
uuid = "0f8b85d8-7281-11e9-16c2-39a750bddbf1"
version = "1.10.0"

[[deps.JuliaVariables]]
deps = ["MLStyle", "NameResolution"]
git-tree-sha1 = "49fb3cb53362ddadb4415e9b73926d6b40709e70"
uuid = "b14d175d-62b4-44ba-8fb7-3064adc8c3ec"
version = "0.2.4"

[[deps.LazyModules]]
git-tree-sha1 = "a560dd966b386ac9ae60bdd3a3d3a326062d3c3e"
uuid = "8cdb02fc-e678-4876-92c5-9defec4f444e"
version = "0.3.1"

[[deps.LibCURL]]
deps = ["LibCURL_jll", "MozillaCACerts_jll"]
uuid = "b27032c2-a3e7-50c8-80cd-2d36dbcbfd21"
version = "0.6.3"

[[deps.LibCURL_jll]]
deps = ["Artifacts", "LibSSH2_jll", "Libdl", "MbedTLS_jll", "Zlib_jll", "nghttp2_jll"]
uuid = "deac9b47-8bc7-5906-a0fe-35ac56dc84c0"
version = "7.84.0+0"

[[deps.LibGit2]]
deps = ["Base64", "NetworkOptions", "Printf", "SHA"]
uuid = "76f85450-5226-5b5a-8eaa-529ad045b433"

[[deps.LibSSH2_jll]]
deps = ["Artifacts", "Libdl", "MbedTLS_jll"]
uuid = "29816b5a-b9ab-546f-933c-edad1886dfa8"
version = "1.10.2+0"

[[deps.Libdl]]
uuid = "8f399da3-3557-5675-b5ff-fb832c97cbdb"

[[deps.Libiconv_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "42b62845d70a619f063a7da093d995ec8e15e778"
uuid = "94ce4f54-9a6c-5748-9c1c-f9c7231a4531"
version = "1.16.1+1"

[[deps.LinearAlgebra]]
deps = ["Libdl", "libblastrampoline_jll"]
uuid = "37e2e46d-f89d-539d-b4ee-838fcccc9c8e"

[[deps.LogExpFunctions]]
deps = ["ChainRulesCore", "ChangesOfVariables", "DocStringExtensions", "InverseFunctions", "IrrationalConstants", "LinearAlgebra"]
git-tree-sha1 = "94d9c52ca447e23eac0c0f074effbcd38830deb5"
uuid = "2ab3a3ac-af41-5b50-aa03-7779005ae688"
version = "0.3.18"

[[deps.Logging]]
uuid = "56ddb016-857b-54e1-b83d-db4d58db5568"

[[deps.LoggingExtras]]
deps = ["Dates", "Logging"]
git-tree-sha1 = "5d4d2d9904227b8bd66386c1138cf4d5ffa826bf"
uuid = "e6f89c97-d47a-5376-807f-9c37f3926c36"
version = "0.4.9"

[[deps.MAT]]
deps = ["BufferedStreams", "CodecZlib", "HDF5", "SparseArrays"]
git-tree-sha1 = "971be550166fe3f604d28715302b58a3f7293160"
uuid = "23992714-dd62-5051-b70f-ba57cb901cac"
version = "0.10.3"

[[deps.MLDatasets]]
deps = ["CSV", "DataDeps", "DataFrames", "DelimitedFiles", "FileIO", "FixedPointNumbers", "GZip", "Glob", "HDF5", "ImageShow", "JLD2", "JSON3", "LazyModules", "MAT", "MLUtils", "NPZ", "Pickle", "Printf", "Requires", "SparseArrays", "Tables"]
git-tree-sha1 = "0def460f055098092043c9533e245ef77b6a8cc8"
uuid = "eb30cadb-4394-5ae3-aed4-317e484a6458"
version = "0.7.5"

[[deps.MLStyle]]
git-tree-sha1 = "0638598b2ea9c60303e036be920df8df60fe2812"
uuid = "d8e11817-5142-5d16-987a-aa16d5891078"
version = "0.4.14"

[[deps.MLUtils]]
deps = ["ChainRulesCore", "DelimitedFiles", "FLoops", "FoldsThreads", "Random", "ShowCases", "Statistics", "StatsBase", "Transducers"]
git-tree-sha1 = "824e9dfc7509cab1ec73ba77b55a916bb2905e26"
uuid = "f1d291b0-491e-4a28-83b9-f70985020b54"
version = "0.2.11"

[[deps.MacroTools]]
deps = ["Markdown", "Random"]
git-tree-sha1 = "42324d08725e200c23d4dfb549e0d5d89dede2d2"
uuid = "1914dd2f-81c6-5fcd-8719-6d5c9610ff09"
version = "0.5.10"

[[deps.MappedArrays]]
git-tree-sha1 = "e8b359ef06ec72e8c030463fe02efe5527ee5142"
uuid = "dbb5928d-eab1-5f90-85c2-b9b0edb7c900"
version = "0.4.1"

[[deps.Markdown]]
deps = ["Base64"]
uuid = "d6f4376e-aef5-505a-96c1-9c027394607a"

[[deps.MbedTLS]]
deps = ["Dates", "MbedTLS_jll", "MozillaCACerts_jll", "Random", "Sockets"]
git-tree-sha1 = "6872f9594ff273da6d13c7c1a1545d5a8c7d0c1c"
uuid = "739be429-bea8-5141-9913-cc70e7f3736d"
version = "1.1.6"

[[deps.MbedTLS_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "c8ffd9c3-330d-5841-b78e-0817d7145fa1"
version = "2.28.0+0"

[[deps.MicroCollections]]
deps = ["BangBang", "InitialValues", "Setfield"]
git-tree-sha1 = "4d5917a26ca33c66c8e5ca3247bd163624d35493"
uuid = "128add7d-3638-4c79-886c-908ea0c25c34"
version = "0.1.3"

[[deps.Missings]]
deps = ["DataAPI"]
git-tree-sha1 = "bf210ce90b6c9eed32d25dbcae1ebc565df2687f"
uuid = "e1d29d7a-bbdc-5cf2-9ac0-f12de2c33e28"
version = "1.0.2"

[[deps.Mmap]]
uuid = "a63ad114-7e13-5084-954f-fe012c677804"

[[deps.MosaicViews]]
deps = ["MappedArrays", "OffsetArrays", "PaddedViews", "StackViews"]
git-tree-sha1 = "b34e3bc3ca7c94914418637cb10cc4d1d80d877d"
uuid = "e94cdb99-869f-56ef-bcf0-1ae2bcbe0389"
version = "0.3.3"

[[deps.MozillaCACerts_jll]]
uuid = "14a3606d-f60d-562e-9121-12d972cd8159"
version = "2022.2.1"

[[deps.NPZ]]
deps = ["Compat", "FileIO", "ZipFile"]
git-tree-sha1 = "45f77b87cb9ed5b519f31e1590258930f3b840ee"
uuid = "15e1cf62-19b3-5cfa-8e77-841668bca605"
version = "0.4.2"

[[deps.NaNMath]]
deps = ["OpenLibm_jll"]
git-tree-sha1 = "a7c3d1da1189a1c2fe843a3bfa04d18d20eb3211"
uuid = "77ba4419-2d1f-58cd-9bb1-8ffee604a2e3"
version = "1.0.1"

[[deps.NameResolution]]
deps = ["PrettyPrint"]
git-tree-sha1 = "1a0fa0e9613f46c9b8c11eee38ebb4f590013c5e"
uuid = "71a1bf82-56d0-4bbc-8a3c-48b961074391"
version = "0.1.5"

[[deps.NetworkOptions]]
uuid = "ca575930-c2e3-43a9-ace4-1e988b2c1908"
version = "1.2.0"

[[deps.OffsetArrays]]
deps = ["Adapt"]
git-tree-sha1 = "f71d8950b724e9ff6110fc948dff5a329f901d64"
uuid = "6fe1bfb0-de20-5000-8ca7-80f57d26f881"
version = "1.12.8"

[[deps.OpenBLAS_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "Libdl"]
uuid = "4536629a-c528-5b80-bd46-f80d51c5b363"
version = "0.3.20+0"

[[deps.OpenLibm_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "05823500-19ac-5b8b-9628-191a04bc5112"
version = "0.8.1+0"

[[deps.OpenSSL]]
deps = ["BitFlags", "Dates", "MozillaCACerts_jll", "OpenSSL_jll", "Sockets"]
git-tree-sha1 = "3c3c4a401d267b04942545b1e964a20279587fd7"
uuid = "4d8831e6-92b7-49fb-bdf8-b643e874388c"
version = "1.3.0"

[[deps.OpenSSL_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "e60321e3f2616584ff98f0a4f18d98ae6f89bbb3"
uuid = "458c3c95-2e84-50aa-8efc-19380b2a3a95"
version = "1.1.17+0"

[[deps.OpenSpecFun_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "13652491f6856acfd2db29360e1bbcd4565d04f1"
uuid = "efe28fd5-8261-553b-a9e1-b2916fc3738e"
version = "0.5.5+0"

[[deps.OrderedCollections]]
git-tree-sha1 = "85f8e6578bf1f9ee0d11e7bb1b1456435479d47c"
uuid = "bac558e1-5e72-5ebc-8fee-abe8a469f55d"
version = "1.4.1"

[[deps.PDMats]]
deps = ["LinearAlgebra", "SparseArrays", "SuiteSparse"]
git-tree-sha1 = "cf494dca75a69712a72b80bc48f59dcf3dea63ec"
uuid = "90014a1f-27ba-587c-ab20-58faa44d9150"
version = "0.11.16"

[[deps.PaddedViews]]
deps = ["OffsetArrays"]
git-tree-sha1 = "03a7a85b76381a3d04c7a1656039197e70eda03d"
uuid = "5432bcbf-9aad-5242-b902-cca2824c8663"
version = "0.5.11"

[[deps.Parsers]]
deps = ["Dates"]
git-tree-sha1 = "6c01a9b494f6d2a9fc180a08b182fcb06f0958a0"
uuid = "69de0a69-1ddd-5017-9359-2bf0b02dc9f0"
version = "2.4.2"

[[deps.Pickle]]
deps = ["DataStructures", "InternedStrings", "Serialization", "SparseArrays", "Strided", "StringEncodings", "ZipFile"]
git-tree-sha1 = "e6a34eb1dc0c498f0774bbfbbbeff2de101f4235"
uuid = "fbb45041-c46e-462f-888f-7c521cafbc2c"
version = "0.3.2"

[[deps.Pkg]]
deps = ["Artifacts", "Dates", "Downloads", "LibGit2", "Libdl", "Logging", "Markdown", "Printf", "REPL", "Random", "SHA", "Serialization", "TOML", "Tar", "UUIDs", "p7zip_jll"]
uuid = "44cfe95a-1eb2-52ea-b672-e2afdf69b78f"
version = "1.8.0"

[[deps.PlutoUI]]
deps = ["AbstractPlutoDingetjes", "Base64", "ColorTypes", "Dates", "Hyperscript", "HypertextLiteral", "IOCapture", "InteractiveUtils", "JSON", "Logging", "Markdown", "Random", "Reexport", "UUIDs"]
git-tree-sha1 = "6e33d318cf8843dade925e35162992145b4eb12f"
uuid = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
version = "0.7.44"

[[deps.PooledArrays]]
deps = ["DataAPI", "Future"]
git-tree-sha1 = "a6062fe4063cdafe78f4a0a81cfffb89721b30e7"
uuid = "2dfb63ee-cc39-5dd5-95bd-886bf059d720"
version = "1.4.2"

[[deps.Preferences]]
deps = ["TOML"]
git-tree-sha1 = "47e5f437cc0e7ef2ce8406ce1e7e24d44915f88d"
uuid = "21216c6a-2e73-6563-6e65-726566657250"
version = "1.3.0"

[[deps.PrettyPrint]]
git-tree-sha1 = "632eb4abab3449ab30c5e1afaa874f0b98b586e4"
uuid = "8162dcfd-2161-5ef2-ae6c-7681170c5f98"
version = "0.2.0"

[[deps.PrettyTables]]
deps = ["Crayons", "Formatting", "Markdown", "Reexport", "Tables"]
git-tree-sha1 = "dfb54c4e414caa595a1f2ed759b160f5a3ddcba5"
uuid = "08abe8d2-0d0c-5749-adfa-8a2ac140af0d"
version = "1.3.1"

[[deps.Printf]]
deps = ["Unicode"]
uuid = "de0858da-6303-5e67-8744-51eddeeeb8d7"

[[deps.QuadGK]]
deps = ["DataStructures", "LinearAlgebra"]
git-tree-sha1 = "3c009334f45dfd546a16a57960a821a1a023d241"
uuid = "1fd47b50-473d-5c70-9696-f719f8f3bcdc"
version = "2.5.0"

[[deps.REPL]]
deps = ["InteractiveUtils", "Markdown", "Sockets", "Unicode"]
uuid = "3fa0cd96-eef1-5676-8a61-b3b8758bbffb"

[[deps.Random]]
deps = ["SHA", "Serialization"]
uuid = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"

[[deps.Reexport]]
git-tree-sha1 = "45e428421666073eab6f2da5c9d310d99bb12f9b"
uuid = "189a3867-3050-52da-a836-e630ba90ab69"
version = "1.2.2"

[[deps.Requires]]
deps = ["UUIDs"]
git-tree-sha1 = "838a3a4188e2ded87a4f9f184b4b0d78a1e91cb7"
uuid = "ae029012-a4dd-5104-9daa-d747884805df"
version = "1.3.0"

[[deps.Rmath]]
deps = ["Random", "Rmath_jll"]
git-tree-sha1 = "bf3188feca147ce108c76ad82c2792c57abe7b1f"
uuid = "79098fc4-a85e-5d69-aa6a-4863f24498fa"
version = "0.7.0"

[[deps.Rmath_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "68db32dff12bb6127bac73c209881191bf0efbb7"
uuid = "f50d1b31-88e8-58de-be2c-1cc44531875f"
version = "0.3.0+0"

[[deps.SHA]]
uuid = "ea8e919c-243c-51af-8825-aaa63cd721ce"
version = "0.7.0"

[[deps.SentinelArrays]]
deps = ["Dates", "Random"]
git-tree-sha1 = "efd23b378ea5f2db53a55ae53d3133de4e080aa9"
uuid = "91c51154-3ec4-41a3-a24f-3f23e20d615c"
version = "1.3.16"

[[deps.Serialization]]
uuid = "9e88b42a-f829-5b0c-bbe9-9e923198166b"

[[deps.Setfield]]
deps = ["ConstructionBase", "Future", "MacroTools", "StaticArraysCore"]
git-tree-sha1 = "e2cc6d8c88613c05e1defb55170bf5ff211fbeac"
uuid = "efcf1570-3423-57d1-acb7-fd33fddbac46"
version = "1.1.1"

[[deps.SharedArrays]]
deps = ["Distributed", "Mmap", "Random", "Serialization"]
uuid = "1a1011a3-84de-559e-8e89-a11a2f7dc383"

[[deps.ShowCases]]
git-tree-sha1 = "7f534ad62ab2bd48591bdeac81994ea8c445e4a5"
uuid = "605ecd9f-84a6-4c9e-81e2-4798472b76a3"
version = "0.1.0"

[[deps.SimpleBufferStream]]
git-tree-sha1 = "874e8867b33a00e784c8a7e4b60afe9e037b74e1"
uuid = "777ac1f9-54b0-4bf8-805c-2214025038e7"
version = "1.1.0"

[[deps.Sockets]]
uuid = "6462fe0b-24de-5631-8697-dd941f90decc"

[[deps.SortingAlgorithms]]
deps = ["DataStructures"]
git-tree-sha1 = "b3363d7460f7d098ca0912c69b082f75625d7508"
uuid = "a2af1166-a08f-5f64-846c-94a0d3cef48c"
version = "1.0.1"

[[deps.SparseArrays]]
deps = ["LinearAlgebra", "Random"]
uuid = "2f01184e-e22b-5df5-ae63-d93ebab69eaf"

[[deps.SpecialFunctions]]
deps = ["ChainRulesCore", "IrrationalConstants", "LogExpFunctions", "OpenLibm_jll", "OpenSpecFun_jll"]
git-tree-sha1 = "d75bda01f8c31ebb72df80a46c88b25d1c79c56d"
uuid = "276daf66-3868-5448-9aa4-cd146d93841b"
version = "2.1.7"

[[deps.SplittablesBase]]
deps = ["Setfield", "Test"]
git-tree-sha1 = "e08a62abc517eb79667d0a29dc08a3b589516bb5"
uuid = "171d559e-b47b-412a-8079-5efa626c420e"
version = "0.1.15"

[[deps.StackViews]]
deps = ["OffsetArrays"]
git-tree-sha1 = "46e589465204cd0c08b4bd97385e4fa79a0c770c"
uuid = "cae243ae-269e-4f55-b966-ac2d0dc13c15"
version = "0.1.1"

[[deps.StaticArraysCore]]
git-tree-sha1 = "6b7ba252635a5eff6a0b0664a41ee140a1c9e72a"
uuid = "1e83bf80-4336-4d27-bf5d-d5a4f845583c"
version = "1.4.0"

[[deps.Statistics]]
deps = ["LinearAlgebra", "SparseArrays"]
uuid = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"

[[deps.StatsAPI]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "f9af7f195fb13589dd2e2d57fdb401717d2eb1f6"
uuid = "82ae8749-77ed-4fe6-ae5f-f523153014b0"
version = "1.5.0"

[[deps.StatsBase]]
deps = ["DataAPI", "DataStructures", "LinearAlgebra", "LogExpFunctions", "Missings", "Printf", "Random", "SortingAlgorithms", "SparseArrays", "Statistics", "StatsAPI"]
git-tree-sha1 = "d1bf48bfcc554a3761a133fe3a9bb01488e06916"
uuid = "2913bbd2-ae8a-5f71-8c99-4fb6c76f3a91"
version = "0.33.21"

[[deps.StatsFuns]]
deps = ["ChainRulesCore", "HypergeometricFunctions", "InverseFunctions", "IrrationalConstants", "LogExpFunctions", "Reexport", "Rmath", "SpecialFunctions"]
git-tree-sha1 = "5783b877201a82fc0014cbf381e7e6eb130473a4"
uuid = "4c63d2b9-4356-54db-8cca-17b64c39e42c"
version = "1.0.1"

[[deps.Strided]]
deps = ["LinearAlgebra", "TupleTools"]
git-tree-sha1 = "a7a664c91104329c88222aa20264e1a05b6ad138"
uuid = "5e0ebb24-38b0-5f93-81fe-25c709ecae67"
version = "1.2.3"

[[deps.StringEncodings]]
deps = ["Libiconv_jll"]
git-tree-sha1 = "50ccd5ddb00d19392577902f0079267a72c5ab04"
uuid = "69024149-9ee7-55f6-a4c4-859efe599b68"
version = "0.3.5"

[[deps.StructTypes]]
deps = ["Dates", "UUIDs"]
git-tree-sha1 = "ca4bccb03acf9faaf4137a9abc1881ed1841aa70"
uuid = "856f2bd8-1eba-4b0a-8007-ebc267875bd4"
version = "1.10.0"

[[deps.SuiteSparse]]
deps = ["Libdl", "LinearAlgebra", "Serialization", "SparseArrays"]
uuid = "4607b0f0-06f3-5cda-b6b1-a6196a1729e9"

[[deps.TOML]]
deps = ["Dates"]
uuid = "fa267f1f-6049-4f14-aa54-33bafae1ed76"
version = "1.0.0"

[[deps.TableTraits]]
deps = ["IteratorInterfaceExtensions"]
git-tree-sha1 = "c06b2f539df1c6efa794486abfb6ed2022561a39"
uuid = "3783bdb8-4a98-5b6b-af9a-565f29a5fe9c"
version = "1.0.1"

[[deps.Tables]]
deps = ["DataAPI", "DataValueInterfaces", "IteratorInterfaceExtensions", "LinearAlgebra", "OrderedCollections", "TableTraits", "Test"]
git-tree-sha1 = "c79322d36826aa2f4fd8ecfa96ddb47b174ac78d"
uuid = "bd369af6-aec1-5ad0-b16a-f7cc5008161c"
version = "1.10.0"

[[deps.Tar]]
deps = ["ArgTools", "SHA"]
uuid = "a4e569a6-e804-4fa4-b0f3-eef7a1d5b13e"
version = "1.10.0"

[[deps.TensorCore]]
deps = ["LinearAlgebra"]
git-tree-sha1 = "1feb45f88d133a655e001435632f019a9a1bcdb6"
uuid = "62fd8b95-f654-4bbd-a8a5-9c27f68ccd50"
version = "0.1.1"

[[deps.Test]]
deps = ["InteractiveUtils", "Logging", "Random", "Serialization"]
uuid = "8dfed614-e22c-5e08-85e1-65c5234f0b40"

[[deps.TranscodingStreams]]
deps = ["Random", "Test"]
git-tree-sha1 = "8a75929dcd3c38611db2f8d08546decb514fcadf"
uuid = "3bb67fe8-82b1-5028-8e26-92a6c54297fa"
version = "0.9.9"

[[deps.Transducers]]
deps = ["Adapt", "ArgCheck", "BangBang", "Baselet", "CompositionsBase", "DefineSingletons", "Distributed", "InitialValues", "Logging", "Markdown", "MicroCollections", "Requires", "Setfield", "SplittablesBase", "Tables"]
git-tree-sha1 = "77fea79baa5b22aeda896a8d9c6445a74500a2c2"
uuid = "28d57a85-8fef-5791-bfe6-a80928e7c999"
version = "0.4.74"

[[deps.Tricks]]
git-tree-sha1 = "6bac775f2d42a611cdfcd1fb217ee719630c4175"
uuid = "410a4b4d-49e4-4fbc-ab6d-cb71b17b3775"
version = "0.1.6"

[[deps.TupleTools]]
git-tree-sha1 = "3c712976c47707ff893cf6ba4354aa14db1d8938"
uuid = "9d95972d-f1c8-5527-a6e0-b4b365fa01f6"
version = "1.3.0"

[[deps.URIs]]
git-tree-sha1 = "e59ecc5a41b000fa94423a578d29290c7266fc10"
uuid = "5c2747f8-b7ea-4ff2-ba2e-563bfd36b1d4"
version = "1.4.0"

[[deps.UUIDs]]
deps = ["Random", "SHA"]
uuid = "cf7118a7-6976-5b1a-9a39-7adc72f591a4"

[[deps.Unicode]]
uuid = "4ec0a83e-493e-50e2-b9ac-8f72acf5a8f5"

[[deps.WeakRefStrings]]
deps = ["DataAPI", "InlineStrings", "Parsers"]
git-tree-sha1 = "b1be2855ed9ed8eac54e5caff2afcdb442d52c23"
uuid = "ea10d353-3f73-51f8-a26c-33c1cb351aa5"
version = "1.4.2"

[[deps.ZipFile]]
deps = ["Libdl", "Printf", "Zlib_jll"]
git-tree-sha1 = "3593e69e469d2111389a9bd06bac1f3d730ac6de"
uuid = "a5390f91-8eb1-5f08-bee0-b1d1ffed6cea"
version = "0.9.4"

[[deps.Zlib_jll]]
deps = ["Libdl"]
uuid = "83775a58-1f1d-513f-b197-d71354ab007a"
version = "1.2.12+3"

[[deps.ZygoteRules]]
deps = ["MacroTools"]
git-tree-sha1 = "8c1a8e4dfacb1fd631745552c8db35d0deb09ea0"
uuid = "700de1a5-db45-46bc-99cf-38207098b444"
version = "0.2.2"

[[deps.libblastrampoline_jll]]
deps = ["Artifacts", "Libdl", "OpenBLAS_jll"]
uuid = "8e850b90-86db-534c-a0d3-1478176c7d93"
version = "5.1.1+0"

[[deps.nghttp2_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "8e850ede-7688-5339-a07c-302acd2aaf8d"
version = "1.48.0+0"

[[deps.p7zip_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "3f19e933-33d8-53b3-aaab-bd5110c3b7a0"
version = "17.4.0+0"
"""

# ╔═╡ Cell order:
# ╟─013e24a0-2787-11ec-1249-5f47f0e4e6a2
# ╠═d99395ed-2bb6-4998-8aed-e288f73b031b
# ╠═d3b72240-b983-4dfa-bbea-c5ada977a4a9
# ╠═4c285c36-a7f4-4c85-8b80-8444fc7e0f30
# ╟─9c1a3c78-8094-4d34-81e4-5b1d68df0d8d
# ╟─eb7624a9-334b-4c57-8fac-9c000a217b99
# ╠═ac6c7a54-5480-458c-8eb6-3fca2c415263
# ╟─69312dd9-4cda-4983-850e-f4f9222b89d5
# ╠═0a308ac8-18b6-44aa-a27e-86358a1a89b5
# ╠═86056b72-ba93-4a53-9943-bc7bc9ceeb12
# ╠═85837bce-cde5-4971-9e28-e9530595246a
# ╠═405886d9-78fe-4fba-9080-6560c86eb113
# ╟─7742ccb2-c30e-4b96-a522-0c9883c35905
# ╠═6f1472e6-8f0b-40f6-baa7-f6ab4dcb489d
# ╠═b5237a83-3c00-4008-8f23-61fb90ca755d
# ╟─a59ccf48-3836-48b5-b28f-ab44d41637d5
# ╠═d6fabf73-d039-48b1-ba02-5de61339b4d2
# ╠═01a43fcd-a999-4533-9826-2d3c52f4d090
# ╠═16564236-60a1-4d92-85f0-441531acd6be
# ╟─fbcb4525-b780-4072-9101-d4e1a3e1d8f6
# ╠═8c115842-4f93-4965-9ba5-9f752d0fa669
# ╠═aa921231-4c0f-48b1-8390-40778de3121c
# ╟─ff194fde-5f61-4cb7-91c5-22a4adede59a
# ╠═816a5d46-1c95-408e-b5a6-bb51bac39ab4
# ╠═7656e820-32b0-42c9-bfac-6b8b6f71ea2d
# ╟─45a2f25e-56e8-4312-a658-6981cdd35fa2
# ╠═02b09245-38f2-4385-b0f1-53cf4136f64e
# ╠═ccfa2549-1b63-43fb-9b0d-e1f96b1543ea
# ╟─17398962-9183-4fa4-8eab-d5d9245f3d00
# ╠═40b3ec56-c62f-42d7-88a1-ca3405a6e598
# ╠═eb49d2cf-e388-431e-a616-45c05893b39d
# ╠═449545d1-9bb7-4a41-b2dc-66ef25d51417
# ╟─36467250-5f49-4149-8856-67b68419c0ec
# ╠═1e6f182f-af6e-4aba-a2db-e9d0f982d6e1
# ╠═a57877b7-529b-477b-b904-d9d77b65397b
# ╟─1a0e23c3-859b-41e4-bdb6-f4db4e4cef44
# ╠═b25dec91-e855-4537-aa49-f43a2fe34e19
# ╟─4fabdb20-bac5-4193-a0d4-73195346ef37
# ╠═393e0f9e-81a0-4f5e-b95c-d63dffbc4deb
# ╠═bb9c5701-15ff-48bb-b4dc-1643fe758ab3
# ╟─d75cf0fd-fd8e-45f7-8d32-b1df812aa407
# ╠═a7a9a2d8-7172-4cb9-9d8a-0ccf4dfe3190
# ╟─8f157646-95b2-4a0f-8497-a51378886b83
# ╠═8c0fbee2-1cde-460f-b914-c6e9f7781fc4
# ╠═56807acf-7530-4fdd-bca3-9ff7cc06ba6e
# ╟─cafe4f1a-8532-4528-9ab7-f6450835d879
# ╠═5bbbc793-61ff-430b-a74b-3f1207e4ebbc
# ╟─909e0c4b-4741-4337-91af-b663c55daee9
# ╠═5df77b3e-91a4-4965-9711-58a9473b3c29
# ╠═2b585960-55f5-4b65-b942-9b5653c39a6c
# ╟─00000000-0000-0000-0000-000000000001
# ╟─00000000-0000-0000-0000-000000000002
